/*
 * Copyright (C) 2005-2006 Micronas USA Inc.
 * Copyright (C) 2009-2010 Sensoray Co., Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (Version 2) as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/unistd.h>
#include <linux/time.h>
#include <linux/vmalloc.h>
#include <linux/pagemap.h>
#include <linux/videodev2.h>
#include <media/v4l2-common.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-subdev.h>
#include <media/videobuf-vmalloc.h>
#include <linux/i2c.h>
#include <linux/mutex.h>
#include <linux/uaccess.h>

#include "compat.h"
#include "go7007.h"


/* MPEG-4 compressed video elementary stream*/
#ifndef V4L2_PIX_FMT_MP4V
#define V4L2_PIX_FMT_MP4V v4l2_fourcc('M', 'P', '4', 'V')
#endif


#define call_all(dev, o, f, args...) \
	v4l2_device_call_until_err(dev, 0, o, f, ##args)


/* const was added to some v4l2_ioctl_ops (see v4l2-ioctl.h) */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0)
#define V4L_CONST_3_10 const
#else
#define V4L_CONST_3_10
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
#define V4L_CONST_3_7 const
#else
#define V4L_CONST_3_7
#endif


struct go7007_buffer {
	struct videobuf_buffer vb;
	int modet_active;
	int frame_offset;
	u8 *buf;
};

struct go7007_file {
	struct videobuf_queue vb_vidq;
};

#if 0 // not used
static u32 get_frame_type_flag(struct go7007_buffer *gobuf, int format)
{
	u8 *f = gobuf->buf;

	switch (format) {
	case GO7007_FORMAT_MJPEG:
		return V4L2_BUF_FLAG_KEYFRAME;
	case GO7007_FORMAT_MPEG4:
		switch ((f[4] >> 6) & 0x3) {
		case 0:
			return V4L2_BUF_FLAG_KEYFRAME;
		case 1:
			return V4L2_BUF_FLAG_PFRAME;
		case 2:
			return V4L2_BUF_FLAG_BFRAME;
		default:
			return 0;
		}
	case GO7007_FORMAT_MPEG1:
	case GO7007_FORMAT_MPEG2:
		switch ((f[5] >> 3) & 0x7) {
		case 1:
			return V4L2_BUF_FLAG_KEYFRAME;
		case 2:
			return V4L2_BUF_FLAG_PFRAME;
		case 3:
			return V4L2_BUF_FLAG_BFRAME;
		default:
			return 0;
		}
	}

	return 0;
}
#endif

/*
 * Store a byte in the current video buffer, if there is one.
 */
static inline void store_byte(struct go7007_buffer *gobuf, u8 byte)
{
	if (gobuf != NULL && gobuf->vb.size < gobuf->vb.bsize)
		gobuf->buf[gobuf->vb.size++] = byte;
}

/*
 * Deliver the last video buffer and get a new one to start writing to.
 */
static struct go7007_buffer *frame_boundary(struct go7007 *go)
{
	struct go7007_buffer *gobuf;
	int i;

	gobuf = go->active_buf;
	if (gobuf) {
		if (gobuf->modet_active) {
			if (gobuf->vb.size + 216 < gobuf->vb.bsize) {
				for (i = 0; i < 216; ++i)
					store_byte(gobuf, go->active_map[i]);
				gobuf->vb.size -= 216;
			} else
				gobuf->modet_active = 0;
		}
		gobuf->vb.state = VIDEOBUF_DONE;
		wake_up(&gobuf->vb.done);
		go->active_buf = NULL;
#if 0
		switch (get_frame_type_flag(gobuf, go->format)) {
		case V4L2_BUF_FLAG_KEYFRAME:
			printk(KERN_INFO "V4L2_BUF_FLAG_KEYFRAME\n");
			break;
		case V4L2_BUF_FLAG_PFRAME:
			printk(KERN_INFO "V4L2_BUF_FLAG_PFRAME\n");
			break;
		case V4L2_BUF_FLAG_BFRAME:
			printk(KERN_INFO "V4L2_BUF_FLAG_BFRAME\n");
			break;
		default:
			printk(KERN_INFO "V4L2_BUF_FLAG_SEQ %02x %02x %02x %02x %02x %02x \n",
			gobuf->buf[0], gobuf->buf[1],
			gobuf->buf[2], gobuf->buf[3],
			gobuf->buf[4], gobuf->buf[5]);
		break;
		}
#endif
	}
	if (!list_empty(&go->vidq.active)) {
		gobuf = list_entry(go->vidq.active.next, 
				   struct go7007_buffer, vb.queue);
		list_del(&gobuf->vb.queue);
		
		if (gobuf->vb.state == VIDEOBUF_QUEUED) {
			gobuf->vb.state = VIDEOBUF_ACTIVE;
			//gobuf->vb.sequence = go->next_seq;
			do_gettimeofday(&gobuf->vb.ts); /* TODO: ktime */
			//ktime_get_ts(&gobuf->vb.ts);
			go->active_buf = gobuf;
			gobuf->vb.size = 0;
			gobuf->buf = videobuf_to_vmalloc(&gobuf->vb);
			BUG_ON(!gobuf->buf);
			gobuf->vb.field_count = go->next_seq << 1;
		}
	}
	++go->next_seq;
	return go->active_buf;
}


static void write_bitmap_word(struct go7007 *go)
{
	int x, y, i, stride = ((go->width >> 4) + 7) >> 3;

	for (i = 0; i < 16; ++i) {
		y = (((go->parse_length - 1) << 3) + i) / (go->width >> 4);
		x = (((go->parse_length - 1) << 3) + i) % (go->width >> 4);
		if (stride * y + (x >> 3) < ARRAY_SIZE(go->active_map))
			go->active_map[stride * y + (x >> 3)] |=
					(go->modet_word & 1) << (x & 0x7);
		go->modet_word >>= 1;
	}
}

/*
 * Parse a chunk of the video stream into frames.  The frames are not
 * delimited by the hardware, so we have to parse the frame boundaries
 * based on the type of video stream we're receiving.
 */
void go7007_parse_video_stream(struct go7007 *go, u8 *buf, int length)
{
	int i, seq_start_code = -1, frame_start_code = -1;
	struct go7007_buffer *gobuf;

	spin_lock(&go->spinlock);

	switch (go->format) {
	case GO7007_FORMAT_MPEG4:
		seq_start_code = 0xB0;
		frame_start_code = 0xB6;
		break;
	case GO7007_FORMAT_MPEG1:
	case GO7007_FORMAT_MPEG2:
		seq_start_code = 0xB3;
		frame_start_code = 0x00;
		break;
	}
	gobuf = go->active_buf;
	
	for (i = 0; i < length; ++i) {
		if (gobuf != NULL && gobuf->vb.size >= gobuf->vb.bsize - 3) {
			v4l2_info(&go->v4l2_dev, "dropping oversized frame\n");
			gobuf->vb.state = VIDEOBUF_QUEUED;
			gobuf->vb.size = 0;
			list_add_tail(&gobuf->vb.queue, &go->vidq.active);
			gobuf = NULL;
			go->active_buf = NULL;
		}

		switch (go->state) {
		case STATE_DATA:
			switch (buf[i]) {
			case 0x00:
				go->state = STATE_00;
				break;
			case 0xFF:
				go->state = STATE_FF;
				break;
			default:
				store_byte(gobuf, buf[i]);
				break;
			}
			break;
		case STATE_00:
			switch (buf[i]) {
			case 0x00:
				go->state = STATE_00_00;
				break;
			case 0xFF:
				store_byte(gobuf, 0x00);
				go->state = STATE_FF;
				break;
			default:
				store_byte(gobuf, 0x00);
				store_byte(gobuf, buf[i]);
				go->state = STATE_DATA;
				break;
			}
			break;
		case STATE_00_00:
			switch (buf[i]) {
			case 0x00:
				store_byte(gobuf, 0x00);
				/* go->state remains STATE_00_00 */
				break;
			case 0x01:
				go->state = STATE_00_00_01;
				break;
			case 0xFF:
				store_byte(gobuf, 0x00);
				store_byte(gobuf, 0x00);
				go->state = STATE_FF;
				break;
			default:
				store_byte(gobuf, 0x00);
				store_byte(gobuf, 0x00);
				store_byte(gobuf, buf[i]);
				go->state = STATE_DATA;
				break;
			}
			break;
		case STATE_00_00_01:
			/* If this is the start of a new MPEG frame,
			 * get a new buffer */
			if ((go->format == GO7007_FORMAT_MPEG1 ||
			     go->format == GO7007_FORMAT_MPEG2 ||
			     go->format == GO7007_FORMAT_MPEG4) &&
			    (buf[i] == seq_start_code ||
			     buf[i] == 0xB8 || /* GOP code */
			     buf[i] == frame_start_code)) {
				if (gobuf == NULL || go->seen_frame)
					gobuf = frame_boundary(go);
				if (buf[i] == frame_start_code) {
					if (gobuf != NULL)
						gobuf->frame_offset =
							gobuf->vb.size;
					go->seen_frame = 1;
				} else {
					go->seen_frame = 0;
				}
			}
			/* Handle any special chunk types, or just write the
			 * start code to the (potentially new) buffer */
			switch (buf[i]) {
			case 0xF5: /* timestamp */
				go->parse_length = 12;
				go->state = STATE_TIMESTAMP;
				break;
			case 0xF6: /* vbi */
				go->state = STATE_VBI_LEN_A;
				break;
			case 0xF8: /* MD map */
				if (go->modet_enable == 0)
					goto modet_not_enabled;
				go->parse_length = 0;
				memset(go->active_map, 0,
						sizeof(go->active_map));
				go->state = STATE_MODET_MAP;
				break;
			case 0xFF: /* Potential JPEG start code */
				store_byte(gobuf, 0x00);
				store_byte(gobuf, 0x00);
				store_byte(gobuf, 0x01);
				go->state = STATE_FF;
				break;
			modet_not_enabled:
			default:
				store_byte(gobuf, 0x00);
				store_byte(gobuf, 0x00);
				store_byte(gobuf, 0x01);
				store_byte(gobuf, buf[i]);
				go->state = STATE_DATA;
				break;
			}
			break;
		case STATE_FF:
			switch (buf[i]) {
			case 0x00:
				store_byte(gobuf, 0xFF);
				go->state = STATE_00;
				break;
			case 0xFF:
				store_byte(gobuf, 0xFF);
				/* go->state remains STATE_FF */
				break;
			case 0xD8:
				if (go->format == GO7007_FORMAT_MJPEG)
					gobuf = frame_boundary(go);
				/* fall through */
			default:
				store_byte(gobuf, 0xFF);
				store_byte(gobuf, buf[i]);
				go->state = STATE_DATA;
				break;
			}
			break;
		case STATE_VBI_LEN_A:
			go->parse_length = buf[i] << 8;
			go->state = STATE_VBI_LEN_B;
			break;
		case STATE_VBI_LEN_B:
			go->parse_length |= buf[i];
			if (go->parse_length > 0)
				go->state = STATE_UNPARSED;
			else
				go->state = STATE_DATA;
			break;
		case STATE_MODET_MAP:
			if (go->parse_length < 204) {
				if (go->parse_length & 1) {
					go->modet_word |= buf[i];
					write_bitmap_word(go);
				} else
					go->modet_word = buf[i] << 8;
			} else if (go->parse_length == 207 && gobuf) {
				gobuf->modet_active = buf[i];
			}
			if (++go->parse_length == 208)
				go->state = STATE_DATA;
			break;
		case STATE_TIMESTAMP: {
			static u8 ts[12];
			ts[12-go->parse_length] = buf[i];
#if 0			
			if (go->parse_length == 1)
				printk(KERN_INFO "timestamp %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x\n",
				ts[0],ts[1],ts[2],ts[3],
				ts[4],ts[5],ts[6],ts[7],
				ts[8],ts[9],ts[10],ts[11]);
#endif
			/* fall thru */
			}
		case STATE_UNPARSED:
			if (--go->parse_length == 0)
				go->state = STATE_DATA;
			break;
		}
	}

	spin_unlock(&go->spinlock);
}
EXPORT_SYMBOL(go7007_parse_video_stream);


/*---------------------------------------------------------
 videobuf operations
 ----------------------------------------------------------*/
static int buffer_setup(struct videobuf_queue *vq, unsigned int *count,
			unsigned int *size)
{
	struct go7007 *go = vq->priv_data;

	*size = go->width * go->height * 2;

	if (0 == *count)
		*count = 2;

	return 0;
}

static void free_buffer(struct videobuf_queue *vq, struct go7007_buffer *buf)
{
	videobuf_vmalloc_free(&buf->vb);
	buf->vb.state = VIDEOBUF_NEEDS_INIT;
}

static int buffer_prepare(struct videobuf_queue *vq, struct videobuf_buffer *vb,
			  enum v4l2_field field)
{
	struct go7007 *go = vq->priv_data;
	struct go7007_buffer *buf = container_of(vb, struct go7007_buffer, vb);
	int rc;

	buf->vb.size = go->width * go->height * 2;

	if (0 != buf->vb.baddr && buf->vb.bsize < buf->vb.size)
		return -EINVAL;

	buf->vb.width = go->width;
	buf->vb.height = go->height;
	buf->vb.field = field;
	if (VIDEOBUF_NEEDS_INIT == buf->vb.state) {
		rc = videobuf_iolock(vq, &buf->vb, NULL);
		if (rc < 0)
			goto fail;
	}

	buf->vb.state = VIDEOBUF_PREPARED;
	return 0;
fail:
	free_buffer(vq, buf);
	return rc;
}

static void buffer_queue(struct videobuf_queue *vq, struct videobuf_buffer *vb)
{
	struct go7007_buffer *buf = container_of(vb, struct go7007_buffer, vb);
	struct go7007 *go = vq->priv_data;
	struct go7007_dmaqueue *vidq = &go->vidq;

	buf->vb.state = VIDEOBUF_QUEUED;
	list_add_tail(&buf->vb.queue, &vidq->active);
	wake_up(&vidq->wait);
}

static void buffer_release(struct videobuf_queue *vq,
			   struct videobuf_buffer *vb)
{
	struct go7007_buffer *buf = container_of(vb, struct go7007_buffer, vb);

	free_buffer(vq, buf);
}

static struct videobuf_queue_ops go7007_video_qops = {
	.buf_setup = buffer_setup,
	.buf_prepare = buffer_prepare,
	.buf_queue = buffer_queue,
	.buf_release = buffer_release,
};

static int vidioc_reqbufs(struct file *file, void *priv,
			  struct v4l2_requestbuffers *req)
{
	struct go7007_file *gofh = priv;
	
	return videobuf_reqbufs(&gofh->vb_vidq, req);
}

static int vidioc_querybuf(struct file *file, void *priv,
			   struct v4l2_buffer *buf)
{
	struct go7007_file *gofh = priv;

	return videobuf_querybuf(&gofh->vb_vidq, buf);
}

static int vidioc_qbuf(struct file *file, void *priv, struct v4l2_buffer *buf)
{
	struct go7007_file *gofh = priv;

	return videobuf_qbuf(&gofh->vb_vidq, buf);
}

static int vidioc_dqbuf(struct file *file, void *priv, struct v4l2_buffer *buf)
{
	struct go7007_file *gofh = priv;
	
	return videobuf_dqbuf(&gofh->vb_vidq, buf, file->f_flags & O_NONBLOCK);
}

#ifdef CONFIG_VIDEO_V4L1_COMPAT
static int vidioc_cgmbuf(struct file *file, void *priv, struct video_mbuf *mbuf)
{
	struct go7007_file *gofh = priv;

	return videobuf_cgmbuf(&gofh->vb_vidq, mbuf, 8);
}
#endif

static int go7007_open(struct file *file)
{
	struct go7007 *go = video_drvdata(file);
	struct go7007_file *gofh;

	if (go->status != STATUS_ONLINE)
		return -EBUSY;
	gofh = kmalloc(sizeof(struct go7007_file), GFP_KERNEL);
	if (gofh == NULL)
		return -ENOMEM;
	++go->ref_count;
	file->private_data = gofh;
	videobuf_queue_vmalloc_init(&gofh->vb_vidq, &go7007_video_qops,
				    NULL, &go->spinlock,
				    V4L2_BUF_TYPE_VIDEO_CAPTURE,
				    V4L2_FIELD_INTERLACED,
				    sizeof(struct go7007_buffer), go
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
					, NULL
#endif
					);
	return 0;
}

static int go7007_release(struct file *file)
{
	struct go7007 *go = video_drvdata(file);
	struct go7007_file *gofh = file->private_data;

	if (go->streaming_file == file) {
		go7007_stream_stop(go);
		go7007_reset_encoder(go);
		go->streaming_file = NULL;
		go->streaming = 0;
	}
	videobuf_stop(&gofh->vb_vidq);
	videobuf_mmap_free(&gofh->vb_vidq);
	
	kfree(gofh);
	if (--go->ref_count == 0)
		kfree(go);
	file->private_data = NULL;
	return 0;
}

static int vidioc_streamon(struct file *file, void *priv,
			   enum v4l2_buf_type type)
{
	struct go7007 *go = video_drvdata(file);
	struct go7007_file *gofh = priv;
	int retval = 0;

	if (type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	mutex_lock(&go->hw_lock);
	if (go->streaming_file != NULL) {
		retval = -EBUSY;
	} else {
		retval = videobuf_streamon(&gofh->vb_vidq);
		if (!go->streaming && retval == 0) {
			go->streaming = 1;
			go->next_seq = 0;
			go->active_buf = NULL;
			if (go7007_start_encoder(go) < 0)
				retval = -EIO;
		}
		if (retval == 0)
			go->streaming_file = file;
	}
	mutex_unlock(&go->hw_lock);

	return retval;
}

static int vidioc_streamoff(struct file *file, void *priv,
			    enum v4l2_buf_type type)
{
	struct go7007 *go = video_drvdata(file);
	struct go7007_file *gofh = priv;

	if (type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	mutex_lock(&go->hw_lock);
	if (go->streaming_file == file) {
		/* Ignoring errors for xawtv sake */
		videobuf_streamoff(&gofh->vb_vidq);
		go7007_stream_stop(go);
		go7007_reset_encoder(go);
		go->streaming_file = NULL;
		go->streaming = 0;
	}
	mutex_unlock(&go->hw_lock);

	return 0;
}

static ssize_t go7007_read(struct file *file, char __user *data,
			   size_t count, loff_t *ppos)
{
	struct go7007 *go = video_drvdata(file);
	struct go7007_file *gofh = file->private_data;
	
	if (go->streaming_file != file) {
		int rc;
		if((rc = vidioc_streamon(file, file->private_data,
				     V4L2_BUF_TYPE_VIDEO_CAPTURE)))
			return rc;
	}
	return videobuf_read_stream(&gofh->vb_vidq, data, count, ppos,
				    0, file->f_flags & O_NONBLOCK);
}

static int go7007_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct go7007_file *gofh = file->private_data;

	return videobuf_mmap_mapper(&gofh->vb_vidq, vma);
}

static unsigned int go7007_poll(struct file *file, poll_table *wait)
{
	struct go7007 *go = video_drvdata(file);
	struct go7007_file *gofh = file->private_data;

	if (go->streaming_file != file) {
		int rc;
		rc = vidioc_streamon(file, file->private_data,
				     V4L2_BUF_TYPE_VIDEO_CAPTURE);
		if (rc)
			return rc;
	}
	return videobuf_poll_stream(file, &gofh->vb_vidq, wait);
}





static int set_capture_size(struct go7007 *go, struct v4l2_format *fmt, int try)
{
	int sensor_height = 0, sensor_width = 0;
	int width, height, i;

	if (fmt != NULL &&
	    fmt->fmt.pix.pixelformat != V4L2_PIX_FMT_MJPEG &&
	    fmt->fmt.pix.pixelformat != V4L2_PIX_FMT_MPEG &&
	    fmt->fmt.pix.pixelformat != V4L2_PIX_FMT_MP4V)
		return -EINVAL;

	switch (go->standard) {
	case GO7007_STD_NTSC:
		sensor_width = 720;
		sensor_height = 480;
		break;
	case GO7007_STD_PAL:
		sensor_width = 720;
		sensor_height = 576;
		break;
	case GO7007_STD_OTHER:
		sensor_width = go->board_info->sensor_width;
		sensor_height = go->board_info->sensor_height;
		break;
	}

	if (fmt == NULL) {
		width = sensor_width;
		height = sensor_height;
	} else if (go->board_info->sensor_flags & GO7007_SENSOR_SCALING) {
		if (fmt->fmt.pix.width > sensor_width)
			width = sensor_width;
		else if (fmt->fmt.pix.width < 144)
			width = 144;
		else
			width = fmt->fmt.pix.width & ~0x0f;

		if (fmt->fmt.pix.height > sensor_height)
			height = sensor_height;
		else if (fmt->fmt.pix.height < 96)
			height = 96;
		else
			height = fmt->fmt.pix.height & ~0x0f;
	} else {
		int requested_size = fmt->fmt.pix.width * fmt->fmt.pix.height;
		int sensor_size = sensor_width * sensor_height;

		if (64 * requested_size < 9 * sensor_size) {
			width = sensor_width / 4;
			height = sensor_height / 4;
		} else if (64 * requested_size < 36 * sensor_size) {
			width = sensor_width / 2;
			height = sensor_height / 2;
		} else {
			width = sensor_width;
			height = sensor_height;
		}
		width &= ~0xf;
		height &= ~0xf;
	}

	if (fmt != NULL) {
		u32 pixelformat = fmt->fmt.pix.pixelformat;

		memset(fmt, 0, sizeof(*fmt));
		fmt->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		fmt->fmt.pix.width = width;
		fmt->fmt.pix.height = height;
		fmt->fmt.pix.pixelformat = pixelformat;
		fmt->fmt.pix.field = V4L2_FIELD_NONE;
		fmt->fmt.pix.bytesperline = width * 2;
		fmt->fmt.pix.sizeimage = fmt->fmt.pix.bytesperline * height;
		fmt->fmt.pix.colorspace = V4L2_COLORSPACE_SMPTE170M; /* ?? */
	}

	if (try)
		return 0;

	go->width = width;
	go->height = height;
	go->encoder_h_offset = go->board_info->sensor_h_offset;
	go->encoder_v_offset = go->board_info->sensor_v_offset;
	for (i = 0; i < 4; ++i)
		go->modet[i].enable = 0;
	for (i = 0; i < 1624; ++i)
		go->modet_map[i] = 0;

	if (go->board_info->sensor_flags & GO7007_SENSOR_SCALING) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 37)
		struct v4l2_format res;

		if (fmt != NULL) {
			res = *fmt;
		} else {
			res.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			res.fmt.pix.width = width;
		}

		if (height > sensor_height / 2) {
			res.fmt.pix.height = height / 2;
			go->encoder_v_halve = 0;
		} else {
			res.fmt.pix.height = height;
			go->encoder_v_halve = 1;
		}
		call_all(&go->v4l2_dev, video, s_fmt, &res);
#else
		struct v4l2_mbus_framefmt mbus_fmt;

		mbus_fmt.code = V4L2_MBUS_FMT_FIXED;
		if (fmt != NULL)
			mbus_fmt.width = fmt->fmt.pix.width;
		else
			mbus_fmt.width = width;

		if (height > sensor_height / 2) {
			mbus_fmt.height = height / 2;
			go->encoder_v_halve = 0;
		} else {
			mbus_fmt.height = height;
			go->encoder_v_halve = 1;
		}
		call_all(&go->v4l2_dev, video, s_mbus_fmt, &mbus_fmt);
#endif
	} else {
		if (width <= sensor_width / 4) {
			go->encoder_h_halve = 1;
			go->encoder_v_halve = 1;
			go->encoder_subsample = 1;
		} else if (width <= sensor_width / 2) {
			go->encoder_h_halve = 1;
			go->encoder_v_halve = 1;
			go->encoder_subsample = 0;
		} else {
			go->encoder_h_halve = 0;
			go->encoder_v_halve = 0;
			go->encoder_subsample = 0;
		}
	}

	if (fmt == NULL)
		return 0;

	switch (fmt->fmt.pix.pixelformat) {
	case V4L2_PIX_FMT_MPEG:
		if (go->format == GO7007_FORMAT_MPEG1 ||
		    go->format == GO7007_FORMAT_MPEG2 ||
		    go->format == GO7007_FORMAT_MPEG4)
			break;
		go->format = GO7007_FORMAT_MPEG1;
		go->pali = 0;
		go->aspect_ratio = GO7007_RATIO_1_1;
		go->gop_size = go->sensor_framerate / 1000;
		go->ipb = 0;
		go->closed_gop = 0;
		go->repeat_seqhead = 0;
		go->seq_header_enable = 1;
		go->gop_header_enable = 1;
		go->dvd_mode = 0;
		break;
	/* Backwards compatibility only! */
	case V4L2_PIX_FMT_MP4V:
		if (go->format == GO7007_FORMAT_MPEG4)
			break;
		go->format = GO7007_FORMAT_MPEG4;
		go->pali = 0xf5;
		go->aspect_ratio = GO7007_RATIO_1_1;
		go->gop_size = go->sensor_framerate / 1000;
		go->ipb = 0;
		go->closed_gop = 0;
		go->repeat_seqhead = 0;
		go->seq_header_enable = 1;
		go->gop_header_enable = 1;
		go->dvd_mode = 0;
		break;
	case V4L2_PIX_FMT_MJPEG:
		go->format = GO7007_FORMAT_MJPEG;
		go->pali = 0;
		go->aspect_ratio = GO7007_RATIO_1_1;
		go->gop_size = 0;
		go->ipb = 0;
		go->closed_gop = 0;
		go->repeat_seqhead = 0;
		go->seq_header_enable = 0;
		go->gop_header_enable = 0;
		go->dvd_mode = 0;
		break;
	}
	return 0;
}

#if 0
static int clip_to_modet_map(struct go7007 *go, int region,
		struct v4l2_clip *clip_list)
{
	struct v4l2_clip clip, *clip_ptr;
	int x, y, mbnum;

	/* Check if coordinates are OK and if any macroblocks are already
	 * used by other regions (besides 0) */
	clip_ptr = clip_list;
	while (clip_ptr) {
		if (copy_from_user(&clip, clip_ptr, sizeof(clip)))
			return -EFAULT;
		if (clip.c.left < 0 || (clip.c.left & 0xF) ||
				clip.c.width <= 0 || (clip.c.width & 0xF))
			return -EINVAL;
		if (clip.c.left + clip.c.width > go->width)
			return -EINVAL;
		if (clip.c.top < 0 || (clip.c.top & 0xF) ||
				clip.c.height <= 0 || (clip.c.height & 0xF))
			return -EINVAL;
		if (clip.c.top + clip.c.height > go->height)
			return -EINVAL;
		for (y = 0; y < clip.c.height; y += 16)
			for (x = 0; x < clip.c.width; x += 16) {
				mbnum = (go->width >> 4) *
						((clip.c.top + y) >> 4) +
					((clip.c.left + x) >> 4);
				if (go->modet_map[mbnum] != 0 &&
						go->modet_map[mbnum] != region)
					return -EBUSY;
			}
		clip_ptr = clip.next;
	}

	/* Clear old region macroblocks */
	for (mbnum = 0; mbnum < 1624; ++mbnum)
		if (go->modet_map[mbnum] == region)
			go->modet_map[mbnum] = 0;

	/* Claim macroblocks in this list */
	clip_ptr = clip_list;
	while (clip_ptr) {
		if (copy_from_user(&clip, clip_ptr, sizeof(clip)))
			return -EFAULT;
		for (y = 0; y < clip.c.height; y += 16)
			for (x = 0; x < clip.c.width; x += 16) {
				mbnum = (go->width >> 4) *
						((clip.c.top + y) >> 4) +
					((clip.c.left + x) >> 4);
				go->modet_map[mbnum] = region;
			}
		clip_ptr = clip.next;
	}
	return 0;
}
#endif


static int vidioc_querycap(struct file *file, void  *priv,
					struct v4l2_capability *cap)
{
	struct go7007 *go = video_drvdata(file);

	strlcpy(cap->driver, "go7007", sizeof(cap->driver));
	strlcpy(cap->card, go->name, sizeof(cap->card));
	strlcpy(cap->bus_info, dev_name(go->dev), sizeof(cap->bus_info));

	cap->version = KERNEL_VERSION(0, 9, 8);

	cap->capabilities = V4L2_CAP_VIDEO_CAPTURE |
			    V4L2_CAP_STREAMING; /* | V4L2_CAP_AUDIO; */

	if (go->board_info->flags & GO7007_BOARD_HAS_TUNER)
		cap->capabilities |= V4L2_CAP_TUNER;

	return 0;
}

static int vidioc_enum_fmt_vid_cap(struct file *file, void  *priv,
					struct v4l2_fmtdesc *fmt)
{
	char *desc = NULL;

	switch (fmt->index) {
	case 0:
		fmt->pixelformat = V4L2_PIX_FMT_MJPEG;
		desc = "Motion-JPEG";
		break;
	case 1:
		fmt->pixelformat = V4L2_PIX_FMT_MPEG;
		desc = "MPEG-1/2/4";
		break;
	case 2:
		fmt->pixelformat = V4L2_PIX_FMT_MP4V;
		desc = "MPEG-4";
		break;
	default:
		return -EINVAL;
	}
	fmt->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt->flags = V4L2_FMT_FLAG_COMPRESSED;

	strncpy(fmt->description, desc, sizeof(fmt->description));

	return 0;
}

static int vidioc_g_fmt_vid_cap(struct file *file, void *priv,
					struct v4l2_format *fmt)
{
	struct go7007 *go = video_drvdata(file);

	fmt->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt->fmt.pix.width = go->width;
	fmt->fmt.pix.height = go->height;
	switch (go->format) {
	case GO7007_FORMAT_MJPEG:
		fmt->fmt.pix.pixelformat = V4L2_PIX_FMT_MJPEG;
		break;
	case GO7007_FORMAT_MPEG4:
		fmt->fmt.pix.pixelformat = V4L2_PIX_FMT_MP4V;
		break;
	default:
		fmt->fmt.pix.pixelformat = V4L2_PIX_FMT_MPEG;
		break;
	}
	fmt->fmt.pix.field = V4L2_FIELD_NONE;
	fmt->fmt.pix.bytesperline = go->width * 2;
	fmt->fmt.pix.sizeimage = fmt->fmt.pix.bytesperline * go->height;
	fmt->fmt.pix.colorspace = V4L2_COLORSPACE_SMPTE170M;

	return 0;
}

static int vidioc_try_fmt_vid_cap(struct file *file, void *priv,
			struct v4l2_format *fmt)
{
	struct go7007 *go = video_drvdata(file);

	return set_capture_size(go, fmt, 1);
}

static int vidioc_s_fmt_vid_cap(struct file *file, void *priv,
			struct v4l2_format *fmt)
{
	struct go7007 *go = video_drvdata(file);

	if (go->streaming)
		return -EBUSY;

	return set_capture_size(go, fmt, 0);
}



/* Controls supported:

contrast, hue, saturation, brightness (via subdev)
seq mode(V4L2_CID_MPEG_VIDEO_B_FRAMES) 1(IONLY), 2(IPONLY)(gop 2-15), 3(IPB)(gop 3,6..15/30)
gop size(V4L2_CID_MPEG_VIDEO_GOP_SIZE) 1-15 (MPEG1/2/4) 1-30 (MPEG4DIVX)
bitrate (V4L2_CID_MPEG_VIDEO_BITRATE) 
encoding (V4L2_CID_MPEG_VIDEO_ENCODING) MPEG1, MPEG2, MPEG4
aspect ratio (V4L2_CID_MPEG_VIDEO_ASPECT)
gop closure (V4L2_CID_MPEG_VIDEO_GOP_CLOSURE)
profile (V4L2_CID_MPEG_VIDEO_PROFILE)

NOTE: subdev controls must use queryctrl, g_ctrl, s_ctrl (not g/s_ext_ctrls)
      and be of class user

*/

static int vidioc_queryctrl(struct file *file, void *priv,
			   struct v4l2_queryctrl *query)
{
	struct go7007 *go = video_drvdata(file);
	static const u32 mpeg_ctrls[] = {
		V4L2_CID_MPEG_CLASS,
		V4L2_CID_MPEG_VIDEO_ENCODING,
		V4L2_CID_MPEG_VIDEO_ASPECT,
		V4L2_CID_MPEG_VIDEO_B_FRAMES,
		V4L2_CID_MPEG_VIDEO_GOP_SIZE,
		V4L2_CID_MPEG_VIDEO_GOP_CLOSURE,
		V4L2_CID_MPEG_VIDEO_BITRATE,
		0
	};
	static const u32 *ctrl_classes[] = {
		mpeg_ctrls,
		NULL
	};
	int id = query->id;

	if (id == V4L2_CTRL_FLAG_NEXT_CTRL)
		id |= V4L2_CID_USER_BASE - 1;

	while (V4L2_CTRL_ID2CLASS(id) == V4L2_CTRL_CLASS_USER) {
		struct v4l2_subdev *sd;

		if (id & V4L2_CTRL_FLAG_NEXT_CTRL)
			id++;
		
		if ((id & V4L2_CTRL_ID_MASK) >= V4L2_CID_LASTP1)
			break;
		
		query->id = id & ~V4L2_CTRL_FLAG_NEXT_CTRL;
		/* query subdevs, returning if one succeeds */
		list_for_each_entry(sd, &go->v4l2_dev.subdevs, list)
                        if (sd->ops->core && sd->ops->core->queryctrl)
                                if (sd->ops->core->queryctrl(sd, query) == 0)
					return 0;
	
		if (!(id & V4L2_CTRL_FLAG_NEXT_CTRL))
			return -EINVAL;
	}
	
	query->id = v4l2_ctrl_next(ctrl_classes, id);
	switch (query->id) {
	case V4L2_CID_MPEG_CLASS:
		return v4l2_ctrl_query_fill(query, 0, 0, 0, 0);
	case V4L2_CID_MPEG_VIDEO_ENCODING:
		return v4l2_ctrl_query_fill(query,
				V4L2_MPEG_VIDEO_ENCODING_MPEG_1,
				V4L2_MPEG_VIDEO_ENCODING_MPEG_4_AVC, 1,
				V4L2_MPEG_VIDEO_ENCODING_MPEG_2);
	case V4L2_CID_MPEG_VIDEO_ASPECT:
		return v4l2_ctrl_query_fill(query,
				V4L2_MPEG_VIDEO_ASPECT_1x1,
				V4L2_MPEG_VIDEO_ASPECT_16x9, 1,
				V4L2_MPEG_VIDEO_ASPECT_1x1);
	case V4L2_CID_MPEG_VIDEO_B_FRAMES:
		return v4l2_ctrl_query_fill(query, 0, 33, 1, 2);
	case V4L2_CID_MPEG_VIDEO_GOP_SIZE:
		return v4l2_ctrl_query_fill(query, 0, 34, 1, 15);
	case V4L2_CID_MPEG_VIDEO_GOP_CLOSURE:
		return v4l2_ctrl_query_fill(query, 0, 1, 1, 1);
	case V4L2_CID_MPEG_VIDEO_BITRATE:
		return v4l2_ctrl_query_fill(query,
				64000,
				10000000, 1,
				1500000);
	default:
		break;
	}
	return -EINVAL;
}

static int vidioc_querymenu(struct file *file, void *priv,
			    struct v4l2_querymenu *qmenu)
{
	switch (qmenu->id) {
	case V4L2_CID_MPEG_VIDEO_ENCODING:
		if (qmenu->index == V4L2_MPEG_VIDEO_ENCODING_MPEG_4_AVC) {
			/* it's really mpeg-4, not AVC */
		    	strlcpy(qmenu->name, "MPEG-4", sizeof(qmenu->name));
			return 0;
		}
		break;
	default:
		break;
	}
	return v4l2_ctrl_query_menu(qmenu, NULL, NULL);
}

static int vidioc_try_ext_ctrls(struct file *file, void *priv,
				struct v4l2_ext_controls *ctrls)
{
	struct v4l2_ext_control *ctrl = ctrls->controls;
	unsigned int i;

	if (ctrls->ctrl_class == V4L2_CTRL_CLASS_USER)
		return 0; /* we assume user controls are ok */
	if (ctrls->ctrl_class != V4L2_CTRL_CLASS_MPEG)
		return -EINVAL;

	/* for invalid controls, clamp or set them to default value */
	for (i = 0; i < ctrls->count; i++, ctrl++) {
		switch(ctrl->id) {
		case V4L2_CID_MPEG_VIDEO_ENCODING:
			if (ctrl->value != V4L2_MPEG_VIDEO_ENCODING_MPEG_1 &&
			    ctrl->value != V4L2_MPEG_VIDEO_ENCODING_MPEG_2 &&
			    ctrl->value != V4L2_MPEG_VIDEO_ENCODING_MPEG_4_AVC)
				ctrl->value = V4L2_MPEG_VIDEO_ENCODING_MPEG_2;
			break;
		case V4L2_CID_MPEG_VIDEO_ASPECT:
			if (ctrl->value != V4L2_MPEG_VIDEO_ASPECT_1x1 &&
			    ctrl->value != V4L2_MPEG_VIDEO_ASPECT_4x3 &&
			    ctrl->value != V4L2_MPEG_VIDEO_ASPECT_16x9)
				ctrl->value = V4L2_MPEG_VIDEO_ASPECT_1x1;
			break;
		case V4L2_CID_MPEG_VIDEO_B_FRAMES:
			if (ctrl->value < 0)
				ctrl->value = 0;
			else if (ctrl->value > 15)
				ctrl->value = 15;
			break;
		case V4L2_CID_MPEG_VIDEO_GOP_SIZE:
			if (ctrl->value < 0)
				ctrl->value = 0;
			else if (ctrl->value > 34)
				ctrl->value = 34;
			break;
		case V4L2_CID_MPEG_VIDEO_GOP_CLOSURE:
			if (ctrl->value != 0)
				ctrl->value = 1;
			break;
		case V4L2_CID_MPEG_VIDEO_BITRATE:
			if (ctrl->value < 64000)
				ctrl->value = 64000;
			else if (ctrl->value > 10000000)
				ctrl->value = 10000000;
			break;
		default:
			ctrls->error_idx = i;
			return -EINVAL;
		}
	}
	return 0;
}

static int vidioc_g_ext_ctrls(struct file *file, void *priv,
			      struct v4l2_ext_controls *ctrls)
{
	struct go7007 *go = video_drvdata(file);
	struct v4l2_ext_control *ctrl = ctrls->controls;
	unsigned int i;
	
	if (ctrls->ctrl_class == V4L2_CTRL_CLASS_USER) {
		for (i = 0; i < ctrls->count; i++, ctrl++) {
			struct v4l2_subdev *sd;
			struct v4l2_control sd_ctrl;
		
			sd_ctrl.id = ctrl->id;
			/* query subdevs, continue if one succeeds */
			list_for_each_entry(sd, &go->v4l2_dev.subdevs, list)
	       	                if (sd->ops->core && sd->ops->core->g_ctrl)
        	                        if (sd->ops->core->g_ctrl(sd, &sd_ctrl)
					    == 0) {
						ctrl->value = sd_ctrl.value;
						sd = NULL;
						break;
					}
			if (sd) {
				/* no subdev succeeded */
				ctrls->error_idx = i;
				return -EINVAL;
			}
		}
		return 0;
	}
	
	if (ctrls->ctrl_class != V4L2_CTRL_CLASS_MPEG)
		return -EINVAL;

	for (i = 0; i < ctrls->count; i++, ctrl++) {
		switch(ctrl->id) {
		case V4L2_CID_MPEG_VIDEO_ENCODING:
			if (go->format == GO7007_FORMAT_MPEG1)
				ctrl->value = V4L2_MPEG_VIDEO_ENCODING_MPEG_1;
			else if (go->format == GO7007_FORMAT_MPEG2)
				ctrl->value = V4L2_MPEG_VIDEO_ENCODING_MPEG_2;
			else if (go->format == GO7007_FORMAT_MPEG4)
				ctrl->value = V4L2_MPEG_VIDEO_ENCODING_MPEG_4_AVC;
			break;
		case V4L2_CID_MPEG_VIDEO_ASPECT:
			if (go->aspect_ratio == GO7007_RATIO_1_1)
				ctrl->value = V4L2_MPEG_VIDEO_ASPECT_1x1;
			else if (go->aspect_ratio == GO7007_RATIO_4_3)
				ctrl->value = V4L2_MPEG_VIDEO_ASPECT_4x3;
			else if (go->aspect_ratio == GO7007_RATIO_16_9)
				ctrl->value = V4L2_MPEG_VIDEO_ASPECT_16x9;
			break;
		case V4L2_CID_MPEG_VIDEO_B_FRAMES:
			ctrl->value = go->ipb ? 2 : 0;
			break;
		case V4L2_CID_MPEG_VIDEO_GOP_SIZE:
			ctrl->value = go->gop_size;
			break;
		case V4L2_CID_MPEG_VIDEO_GOP_CLOSURE:
			ctrl->value = go->closed_gop;
			break;
		case V4L2_CID_MPEG_VIDEO_BITRATE:
			ctrl->value = go->bitrate;
			break;
		default:
			ctrls->error_idx = i;
			return -EINVAL;
		}
	}
	return 0;
}

static int vidioc_s_ext_ctrls(struct file *file, void *priv,
			      struct v4l2_ext_controls *ctrls)
{
	struct go7007 *go = video_drvdata(file);
	struct v4l2_ext_control *ctrl = ctrls->controls;
	unsigned int i;
	
	if (ctrls->ctrl_class == V4L2_CTRL_CLASS_USER) {
		for (i = 0; i < ctrls->count; i++, ctrl++) {
			struct v4l2_subdev *sd;
			struct v4l2_control sd_ctrl;

			sd_ctrl.id = ctrl->id;
			sd_ctrl.value = ctrl->value;
			/* query subdevs, continue until one succeeds */
			list_for_each_entry(sd, &go->v4l2_dev.subdevs, list)
	       	                if (sd->ops->core && sd->ops->core->g_ctrl)
        	                        if (sd->ops->core->g_ctrl(sd, &sd_ctrl)
					    == 0) {
						sd = NULL;
						break;
					}
			if (sd) {
				/* no subdev succeeded */
				ctrls->error_idx = i;
				return -EINVAL;
			}
		}
		return 0;
	}

	if (vidioc_try_ext_ctrls(file, priv, ctrls))
		return -EINVAL;

	for (i = 0; i < ctrls->count; i++, ctrl++) {
		switch(ctrl->id) {
		case V4L2_CID_MPEG_VIDEO_ENCODING:
			if (ctrl->value == V4L2_MPEG_VIDEO_ENCODING_MPEG_1) {
				go->format = GO7007_FORMAT_MPEG1;
				go->pali = 0;
			} else if (ctrl->value == V4L2_MPEG_VIDEO_ENCODING_MPEG_2) {
				go->format = GO7007_FORMAT_MPEG2;
				go->pali = GO7007_MPEG2_PROFILE_MAIN_MAIN & 0xff;
			} else if (ctrl->value == V4L2_MPEG_VIDEO_ENCODING_MPEG_4_AVC) {
				go->format = GO7007_FORMAT_MPEG4;
				go->pali = GO7007_MPEG4_PROFILE_AS_L5 & 0xff;
			}
			break;
		case V4L2_CID_MPEG_VIDEO_ASPECT:
			if (ctrl->value == V4L2_MPEG_VIDEO_ASPECT_1x1)
				go->aspect_ratio = GO7007_RATIO_1_1;
			else if (ctrl->value == V4L2_MPEG_VIDEO_ASPECT_4x3)
				go->aspect_ratio = GO7007_RATIO_4_3;
			else if (ctrl->value == V4L2_MPEG_VIDEO_ASPECT_16x9)
				go->aspect_ratio = GO7007_RATIO_16_9;
			break;
		case V4L2_CID_MPEG_VIDEO_B_FRAMES:
			go->ipb = ctrl->value ? 1 : 0;
			break;
		case V4L2_CID_MPEG_VIDEO_GOP_SIZE:
			go->gop_size = ctrl->value;
			break;
		case V4L2_CID_MPEG_VIDEO_GOP_CLOSURE:
			go->closed_gop = ctrl->value;
			break;
		case V4L2_CID_MPEG_VIDEO_BITRATE:
			go->bitrate = ctrl->value;
			break;
		default:
			ctrls->error_idx = i;
			return -EINVAL;
		}
	}
	return 0;
}


static int vidioc_g_parm(struct file *file, void *priv,
		struct v4l2_streamparm *parm)
{
	struct go7007 *go = video_drvdata(file);
	struct v4l2_fract timeperframe = {
		.numerator = 1001 *  go->fps_scale,
		.denominator = go->sensor_framerate,
	};

	if (parm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	parm->parm.capture.capability |= V4L2_CAP_TIMEPERFRAME;
	parm->parm.capture.timeperframe = timeperframe;

	return 0;
}

static int vidioc_s_parm(struct file *file, void *priv,
		struct v4l2_streamparm *parm)
{
	struct go7007 *go = video_drvdata(file);
	unsigned int n, d;

	if (parm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;
	if (parm->parm.capture.capturemode != 0)
		return -EINVAL;

	n = go->sensor_framerate *
		parm->parm.capture.timeperframe.numerator;
	d = 1001 * parm->parm.capture.timeperframe.denominator;
	if (n != 0 && d != 0 && n > d)
		go->fps_scale = (n + d/2) / d;
	else
		go->fps_scale = 1;

	return 0;
}

/* VIDIOC_ENUMSTD on go7007 were used for enumberating the supported fps and
   its resolution, when the device is not connected to TV.
   This were an API abuse, probably used by the lack of specific IOCTL's to
   enumberate it, by the time the driver were written.

   However, since kernel 2.6.19, two new ioctls (VIDIOC_ENUM_FRAMEINTERVALS
   and VIDIOC_ENUM_FRAMESIZES) were added for this purpose.

   The two functions bellow implements the newer ioctls
*/
static int vidioc_enum_framesizes(struct file *file, void *priv,
				  struct v4l2_frmsizeenum *fsize)
{
	struct go7007 *go = video_drvdata(file);

	/* Return -EINVAL, if it is a TV board */
	if ((go->board_info->flags & GO7007_BOARD_HAS_TUNER) ||
	    (go->board_info->sensor_flags & GO7007_SENSOR_TV))
		return -EINVAL;

	if (fsize->index > 0)
		return -EINVAL;

	fsize->type = V4L2_FRMSIZE_TYPE_DISCRETE;
	fsize->discrete.width = go->board_info->sensor_width;
	fsize->discrete.height = go->board_info->sensor_height;

	return 0;
}

static int vidioc_enum_frameintervals(struct file *file, void *priv,
				      struct v4l2_frmivalenum *fival)
{
	struct go7007 *go = video_drvdata(file);

	/* Return -EINVAL, if it is a TV board */
	if ((go->board_info->flags & GO7007_BOARD_HAS_TUNER) ||
	    (go->board_info->sensor_flags & GO7007_SENSOR_TV))
		return -EINVAL;

	if (fival->index > 0)
		return -EINVAL;

	fival->type = V4L2_FRMIVAL_TYPE_DISCRETE;
	fival->discrete.numerator = 1001;
	fival->discrete.denominator = go->board_info->sensor_framerate;

	return 0;
}

static int vidioc_g_std(struct file *file, void *priv, v4l2_std_id *std)
{
	struct go7007 *go = video_drvdata(file);

	switch (go->standard) {
	case GO7007_STD_NTSC:
		*std = V4L2_STD_NTSC;
		break;
	case GO7007_STD_PAL:
		*std = V4L2_STD_PAL;
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0)
static int vidioc_s_std(struct file *file, void *priv, v4l2_std_id _std)
{
	v4l2_std_id *std = &_std;
#else
static int vidioc_s_std(struct file *file, void *priv, v4l2_std_id *std)
{
#endif
	struct go7007 *go = video_drvdata(file);

	if (!(go->board_info->sensor_flags & GO7007_SENSOR_TV) && *std != 0)
		return -EINVAL;

	if (*std == 0)
		return -EINVAL;

	if ((go->board_info->flags & GO7007_BOARD_HAS_TUNER) &&
			go->input == go->board_info->num_inputs - 1) {
		if (!go->i2c_adapter_online)
			return -EIO;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
		if (call_all(&go->v4l2_dev, core, s_std, *std) < 0)
#else
		if (call_all(&go->v4l2_dev, tuner, s_std, *std) < 0)
#endif
			return -EINVAL;
	}

	if (*std & V4L2_STD_NTSC) {
		go->standard = GO7007_STD_NTSC;
		go->sensor_framerate = 30000;
	} else if (*std & V4L2_STD_PAL) {
		go->standard = GO7007_STD_PAL;
		go->sensor_framerate = 25025;
	} else if (*std & V4L2_STD_SECAM) {
		go->standard = GO7007_STD_PAL;
		go->sensor_framerate = 25025;
	} else
		return -EINVAL;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
	call_all(&go->v4l2_dev, core, s_std, *std);
#else
	call_all(&go->v4l2_dev, video, s_std_output, *std);
#endif
	set_capture_size(go, NULL, 0);

	return 0;
}

static int vidioc_querystd(struct file *file, void *priv, v4l2_std_id *std)
{
	struct go7007 *go = video_drvdata(file);

	if ((go->board_info->flags & GO7007_BOARD_HAS_TUNER) &&
			go->input == go->board_info->num_inputs - 1) {
		if (!go->i2c_adapter_online)
			return -EIO;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
		return call_all(&go->v4l2_dev, video, querystd, std);
#else
		*std = V4L2_STD_NTSC | V4L2_STD_PAL | V4L2_STD_SECAM;
#endif		
	} else if (go->board_info->sensor_flags & GO7007_SENSOR_TV)
		*std = V4L2_STD_NTSC | V4L2_STD_PAL | V4L2_STD_SECAM;
	else
		*std = 0;

	return 0;
}

static int vidioc_enum_input(struct file *file, void *priv,
				struct v4l2_input *inp)
{
	struct go7007 *go = video_drvdata(file);

	if (inp->index >= go->board_info->num_inputs)
		return -EINVAL;

	strncpy(inp->name, go->board_info->inputs[inp->index].name,
			sizeof(inp->name));

	/* If this board has a tuner, it will be the last input */
	if ((go->board_info->flags & GO7007_BOARD_HAS_TUNER) &&
			inp->index == go->board_info->num_inputs - 1)
		inp->type = V4L2_INPUT_TYPE_TUNER;
	else
		inp->type = V4L2_INPUT_TYPE_CAMERA;

	inp->audioset = 0;
	inp->tuner = 0;
	if (go->board_info->sensor_flags & GO7007_SENSOR_TV)
		inp->std = V4L2_STD_NTSC | V4L2_STD_PAL |
						V4L2_STD_SECAM;
	else
		inp->std = 0;

	return 0;
}


static int vidioc_g_input(struct file *file, void *priv, unsigned int *input)
{
	struct go7007 *go = video_drvdata(file);

	*input = go->input;

	return 0;
}

static int vidioc_s_input(struct file *file, void *priv, unsigned int input)
{
	struct go7007 *go = video_drvdata(file);

	if (input >= go->board_info->num_inputs)
		return -EINVAL;
	if (go->streaming)
		return -EBUSY;

	go->input = input;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 30)
	return call_all(&go->v4l2_dev, video, s_routing, input, 0, 0);
#else
	do {
		struct v4l2_routing routing = { .input = input };
		return call_all(&go->v4l2_dev, video, s_routing, &routing);
	} while (0);
#endif
}

static int vidioc_g_tuner(struct file *file, void *priv,
				struct v4l2_tuner *t)
{
	struct go7007 *go = video_drvdata(file);

	if (!(go->board_info->flags & GO7007_BOARD_HAS_TUNER))
		return -EINVAL;
	if (t->index != 0)
		return -EINVAL;
	if (!go->i2c_adapter_online)
		return -EIO;

	return call_all(&go->v4l2_dev, tuner, g_tuner, t);
}

static int vidioc_s_tuner(struct file *file, void *priv,
				V4L_CONST_3_10 struct v4l2_tuner *t)
{
	struct go7007 *go = video_drvdata(file);

	if (!(go->board_info->flags & GO7007_BOARD_HAS_TUNER))
		return -EINVAL;
	if (t->index != 0)
		return -EINVAL;
	if (!go->i2c_adapter_online)
		return -EIO;

	switch (go->board_id) {
	case GO7007_BOARDID_PX_TV402U_NA:
	case GO7007_BOARDID_PX_TV402U_JP:
		/* No selectable options currently */
		if (t->audmode != V4L2_TUNER_MODE_STEREO)
			return -EINVAL;
		break;
	}

	return call_all(&go->v4l2_dev, tuner, s_tuner, t);
}

static int vidioc_g_frequency(struct file *file, void *priv,
				struct v4l2_frequency *f)
{
	struct go7007 *go = video_drvdata(file);

	if (!(go->board_info->flags & GO7007_BOARD_HAS_TUNER))
		return -EINVAL;
	if (!go->i2c_adapter_online)
		return -EIO;

	f->type = V4L2_TUNER_ANALOG_TV;

	return call_all(&go->v4l2_dev, tuner, g_frequency, f);
}

static int vidioc_s_frequency(struct file *file, void *priv,
				V4L_CONST_3_10 struct v4l2_frequency *f)
{
	struct go7007 *go = video_drvdata(file);

	if (!(go->board_info->flags & GO7007_BOARD_HAS_TUNER))
		return -EINVAL;
	if (!go->i2c_adapter_online)
		return -EIO;

	return call_all(&go->v4l2_dev, tuner, s_frequency, f);
}

static int vidioc_cropcap(struct file *file, void *priv,
					struct v4l2_cropcap *cropcap)
{
	struct go7007 *go = video_drvdata(file);

	if (cropcap->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	/* These specify the raw input of the sensor */
	switch (go->standard) {
	case GO7007_STD_NTSC:
		cropcap->bounds.top = 0;
		cropcap->bounds.left = 0;
		cropcap->bounds.width = 720;
		cropcap->bounds.height = 480;
		cropcap->defrect.top = 0;
		cropcap->defrect.left = 0;
		cropcap->defrect.width = 720;
		cropcap->defrect.height = 480;
		break;
	case GO7007_STD_PAL:
		cropcap->bounds.top = 0;
		cropcap->bounds.left = 0;
		cropcap->bounds.width = 720;
		cropcap->bounds.height = 576;
		cropcap->defrect.top = 0;
		cropcap->defrect.left = 0;
		cropcap->defrect.width = 720;
		cropcap->defrect.height = 576;
		break;
	case GO7007_STD_OTHER:
		cropcap->bounds.top = 0;
		cropcap->bounds.left = 0;
		cropcap->bounds.width = go->board_info->sensor_width;
		cropcap->bounds.height = go->board_info->sensor_height;
		cropcap->defrect.top = 0;
		cropcap->defrect.left = 0;
		cropcap->defrect.width = go->board_info->sensor_width;
		cropcap->defrect.height = go->board_info->sensor_height;
		break;
	}

	return 0;
}

static int vidioc_g_crop(struct file *file, void *priv, struct v4l2_crop *crop)
{
	struct go7007 *go = video_drvdata(file);

	if (crop->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	crop->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	/* These specify the raw input of the sensor */
	switch (go->standard) {
	case GO7007_STD_NTSC:
		crop->c.top = 0;
		crop->c.left = 0;
		crop->c.width = 720;
		crop->c.height = 480;
		break;
	case GO7007_STD_PAL:
		crop->c.top = 0;
		crop->c.left = 0;
		crop->c.width = 720;
		crop->c.height = 576;
		break;
	case GO7007_STD_OTHER:
		crop->c.top = 0;
		crop->c.left = 0;
		crop->c.width = go->board_info->sensor_width;
		crop->c.height = go->board_info->sensor_height;
		break;
	}

	return 0;
}

/* FIXME: vidioc_s_crop is not really implemented!!!
 */
static int vidioc_s_crop(struct file *file, void *priv, V4L_CONST_3_7 struct v4l2_crop *crop)
{
	if (crop->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;

	return 0;
}

static int vidioc_g_jpegcomp(struct file *file, void *priv,
			 struct v4l2_jpegcompression *params)
{
	memset(params, 0, sizeof(*params));
	params->quality = 50; /* ?? */
	params->jpeg_markers = V4L2_JPEG_MARKER_DHT |
				V4L2_JPEG_MARKER_DQT;

	return 0;
}

static int vidioc_s_jpegcomp(struct file *file, void *priv,
			 V4L_CONST_3_7 struct v4l2_jpegcompression *params)
{
	if (params->quality != 50 ||
			params->jpeg_markers != (V4L2_JPEG_MARKER_DHT |
						V4L2_JPEG_MARKER_DQT))
		return -EINVAL;

	return 0;
}

/* FIXME:
	Those ioctls are private, and not needed, since several standard
	extended controls already provide streaming control.
	So, those ioctls should be converted into vidioc_g_ext_ctrls()
	and vidioc_s_ext_ctrls()
 */

#if 0
	/* Temporary ioctls for controlling compression characteristics */
	case GO7007IOC_S_BITRATE:
	{
		int *bitrate = arg;

		if (go->streaming)
			return -EINVAL;
		/* Upper bound is kind of arbitrary here */
		if (*bitrate < 64000 || *bitrate > 10000000)
			return -EINVAL;
		go->bitrate = *bitrate;
		return 0;
	}
	case GO7007IOC_G_BITRATE:
	{
		int *bitrate = arg;

		*bitrate = go->bitrate;
		return 0;
	}
	case GO7007IOC_S_COMP_PARAMS:
	{
		struct go7007_comp_params *comp = arg;

		if (go->format == GO7007_FORMAT_MJPEG)
			return -EINVAL;
		if (comp->gop_size > 0)
			go->gop_size = comp->gop_size;
		else
			go->gop_size = go->sensor_framerate / 1000;
		if (go->gop_size != 15)
			go->dvd_mode = 0;
		/*go->ipb = comp->max_b_frames > 0;*/ /* completely untested */
		if (go->board_info->sensor_flags & GO7007_SENSOR_TV) {
			switch (comp->aspect_ratio) {
			case GO7007_ASPECT_RATIO_4_3_NTSC:
			case GO7007_ASPECT_RATIO_4_3_PAL:
				go->aspect_ratio = GO7007_RATIO_4_3;
				break;
			case GO7007_ASPECT_RATIO_16_9_NTSC:
			case GO7007_ASPECT_RATIO_16_9_PAL:
				go->aspect_ratio = GO7007_RATIO_16_9;
				break;
			default:
				go->aspect_ratio = GO7007_RATIO_1_1;
				break;
			}
		}
		if (comp->flags & GO7007_COMP_OMIT_SEQ_HEADER) {
			go->dvd_mode = 0;
			go->seq_header_enable = 0;
		} else {
			go->seq_header_enable = 1;
		}
		/* fall-through */
	}
	case GO7007IOC_G_COMP_PARAMS:
	{
		struct go7007_comp_params *comp = arg;

		if (go->format == GO7007_FORMAT_MJPEG)
			return -EINVAL;
		memset(comp, 0, sizeof(*comp));
		comp->gop_size = go->gop_size;
		comp->max_b_frames = go->ipb ? 2 : 0;
		switch (go->aspect_ratio) {
		case GO7007_RATIO_4_3:
			if (go->standard == GO7007_STD_NTSC)
				comp->aspect_ratio =
					GO7007_ASPECT_RATIO_4_3_NTSC;
			else
				comp->aspect_ratio =
					GO7007_ASPECT_RATIO_4_3_PAL;
			break;
		case GO7007_RATIO_16_9:
			if (go->standard == GO7007_STD_NTSC)
				comp->aspect_ratio =
					GO7007_ASPECT_RATIO_16_9_NTSC;
			else
				comp->aspect_ratio =
					GO7007_ASPECT_RATIO_16_9_PAL;
			break;
		default:
			comp->aspect_ratio = GO7007_ASPECT_RATIO_1_1;
			break;
		}
		if (go->closed_gop)
			comp->flags |= GO7007_COMP_CLOSED_GOP;
		if (!go->seq_header_enable)
			comp->flags |= GO7007_COMP_OMIT_SEQ_HEADER;
		return 0;
	}
	case GO7007IOC_S_MPEG_PARAMS:
	{
		struct go7007_mpeg_params *mpeg = arg;

		if (go->format != GO7007_FORMAT_MPEG1 &&
				go->format != GO7007_FORMAT_MPEG2 &&
				go->format != GO7007_FORMAT_MPEG4)
			return -EINVAL;

		if (mpeg->flags & GO7007_MPEG_FORCE_DVD_MODE) {
			go->format = GO7007_FORMAT_MPEG2;
			go->bitrate = 9800000;
			go->gop_size = 15;
			go->pali = 0x48;
			go->closed_gop = 1;
			go->repeat_seqhead = 0;
			go->seq_header_enable = 1;
			go->gop_header_enable = 1;
			go->dvd_mode = 1;
		} else {
			switch (mpeg->mpeg_video_standard) {
			case GO7007_MPEG_VIDEO_MPEG1:
				go->format = GO7007_FORMAT_MPEG1;
				go->pali = 0;
				break;
			case GO7007_MPEG_VIDEO_MPEG2:
				go->format = GO7007_FORMAT_MPEG2;
				if (mpeg->pali >> 24 == 2)
					go->pali = mpeg->pali & 0xff;
				else
					go->pali = 0x48;
				break;
			case GO7007_MPEG_VIDEO_MPEG4:
				go->format = GO7007_FORMAT_MPEG4;
				if (mpeg->pali >> 24 == 4)
					go->pali = mpeg->pali & 0xff;
				else
					go->pali = 0xf5;
				break;
			default:
				return -EINVAL;
			}
			go->gop_header_enable =
				mpeg->flags & GO7007_MPEG_OMIT_GOP_HEADER
				? 0 : 1;
			if (mpeg->flags & GO7007_MPEG_REPEAT_SEQHEADER)
				go->repeat_seqhead = 1;
			else
				go->repeat_seqhead = 0;
			go->dvd_mode = 0;
		}
		/* fall-through */
	}
	case GO7007IOC_G_MPEG_PARAMS:
	{
		struct go7007_mpeg_params *mpeg = arg;

		memset(mpeg, 0, sizeof(*mpeg));
		switch (go->format) {
		case GO7007_FORMAT_MPEG1:
			mpeg->mpeg_video_standard = GO7007_MPEG_VIDEO_MPEG1;
			mpeg->pali = 0;
			break;
		case GO7007_FORMAT_MPEG2:
			mpeg->mpeg_video_standard = GO7007_MPEG_VIDEO_MPEG2;
			mpeg->pali = GO7007_MPEG_PROFILE(2, go->pali);
			break;
		case GO7007_FORMAT_MPEG4:
			mpeg->mpeg_video_standard = GO7007_MPEG_VIDEO_MPEG4;
			mpeg->pali = GO7007_MPEG_PROFILE(4, go->pali);
			break;
		default:
			return -EINVAL;
		}
		if (!go->gop_header_enable)
			mpeg->flags |= GO7007_MPEG_OMIT_GOP_HEADER;
		if (go->repeat_seqhead)
			mpeg->flags |= GO7007_MPEG_REPEAT_SEQHEADER;
		if (go->dvd_mode)
			mpeg->flags |= GO7007_MPEG_FORCE_DVD_MODE;
		return 0;
	}
	case GO7007IOC_S_MD_PARAMS:
	{
		struct go7007_md_params *mdp = arg;

		if (mdp->region > 3)
			return -EINVAL;
		if (mdp->trigger > 0) {
			go->modet[mdp->region].pixel_threshold =
					mdp->pixel_threshold >> 1;
			go->modet[mdp->region].motion_threshold =
					mdp->motion_threshold >> 1;
			go->modet[mdp->region].mb_threshold =
					mdp->trigger >> 1;
			go->modet[mdp->region].enable = 1;
		} else
			go->modet[mdp->region].enable = 0;
		/* fall-through */
	}
	case GO7007IOC_G_MD_PARAMS:
	{
		struct go7007_md_params *mdp = arg;
		int region = mdp->region;

		if (mdp->region > 3)
			return -EINVAL;
		memset(mdp, 0, sizeof(struct go7007_md_params));
		mdp->region = region;
		if (!go->modet[region].enable)
			return 0;
		mdp->pixel_threshold =
			(go->modet[region].pixel_threshold << 1) + 1;
		mdp->motion_threshold =
			(go->modet[region].motion_threshold << 1) + 1;
		mdp->trigger =
			(go->modet[region].mb_threshold << 1) + 1;
		return 0;
	}
	case GO7007IOC_S_MD_REGION:
	{
		struct go7007_md_region *region = arg;

		if (region->region < 1 || region->region > 3)
			return -EINVAL;
		return clip_to_modet_map(go, region->region, region->clips);
	}
#endif

static void go7007_vfl_release(struct video_device *vfd)
{
	struct go7007 *go = video_get_drvdata(vfd);

	video_device_release(vfd);
	if (--go->ref_count == 0)
		kfree(go);
}

static struct v4l2_file_operations go7007_fops = {
	.owner		= THIS_MODULE,
	.open		= go7007_open,
	.release	= go7007_release,
	.ioctl		= video_ioctl2,
	.read		= go7007_read,
	.mmap		= go7007_mmap,
	.poll		= go7007_poll,
};

static const struct v4l2_ioctl_ops video_ioctl_ops = {
	.vidioc_querycap          = vidioc_querycap,
	.vidioc_enum_fmt_vid_cap  = vidioc_enum_fmt_vid_cap,
	.vidioc_g_fmt_vid_cap     = vidioc_g_fmt_vid_cap,
	.vidioc_try_fmt_vid_cap   = vidioc_try_fmt_vid_cap,
	.vidioc_s_fmt_vid_cap     = vidioc_s_fmt_vid_cap,
	.vidioc_reqbufs           = vidioc_reqbufs,
	.vidioc_querybuf          = vidioc_querybuf,
	.vidioc_qbuf              = vidioc_qbuf,
	.vidioc_dqbuf             = vidioc_dqbuf,
	.vidioc_g_std             = vidioc_g_std,
	.vidioc_s_std             = vidioc_s_std,
	.vidioc_querystd          = vidioc_querystd,
	.vidioc_enum_input        = vidioc_enum_input,
	.vidioc_g_input           = vidioc_g_input,
	.vidioc_s_input           = vidioc_s_input,
	.vidioc_queryctrl         = vidioc_queryctrl,
	.vidioc_querymenu         = vidioc_querymenu,
	.vidioc_try_ext_ctrls     = vidioc_try_ext_ctrls,
	.vidioc_g_ext_ctrls       = vidioc_g_ext_ctrls,
	.vidioc_s_ext_ctrls	  = vidioc_s_ext_ctrls,	
	.vidioc_streamon          = vidioc_streamon,
	.vidioc_streamoff         = vidioc_streamoff,
	.vidioc_g_tuner           = vidioc_g_tuner,
	.vidioc_s_tuner           = vidioc_s_tuner,
	.vidioc_g_frequency       = vidioc_g_frequency,
	.vidioc_s_frequency       = vidioc_s_frequency,
	.vidioc_g_parm            = vidioc_g_parm,
	.vidioc_s_parm            = vidioc_s_parm,
	.vidioc_enum_framesizes   = vidioc_enum_framesizes,
	.vidioc_enum_frameintervals = vidioc_enum_frameintervals,
	.vidioc_cropcap           = vidioc_cropcap,
	.vidioc_g_crop            = vidioc_g_crop,
	.vidioc_s_crop            = vidioc_s_crop,
	.vidioc_g_jpegcomp        = vidioc_g_jpegcomp,
	.vidioc_s_jpegcomp        = vidioc_s_jpegcomp,
#ifdef CONFIG_VIDEO_V4L1_COMPAT
	.vidiocgmbuf              = vidioc_cgmbuf,
#endif
};

static struct video_device go7007_template = {
	.name		= "go7007",
	.fops		= &go7007_fops,
	.release	= go7007_vfl_release,
	.ioctl_ops	= &video_ioctl_ops,
	.tvnorms	= V4L2_STD_ALL,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 11, 0)
	.current_norm	= V4L2_STD_NTSC,
#endif
};

int go7007_v4l2_init(struct go7007 *go)
{
	int rv;

	go->video_dev = video_device_alloc();
	if (go->video_dev == NULL)
		return -ENOMEM;
	*go->video_dev = go7007_template;
	//go->video_dev->debug = 99;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 11, 0)
	go->video_dev->dev_parent = go->dev;
#else
	go->video_dev->parent = go->dev;
#endif
	rv = video_register_device(go->video_dev, VFL_TYPE_GRABBER, -1);
	if (rv < 0) {
		video_device_release(go->video_dev);
		go->video_dev = NULL;
		return rv;
	}
	rv = v4l2_device_register(go->dev, &go->v4l2_dev);
	if (rv < 0) {
		video_device_release(go->video_dev);
		go->video_dev = NULL;
		return rv;
	}
	video_set_drvdata(go->video_dev, go);
	++go->ref_count;
	printk(KERN_INFO "%s: registered device %s [v4l2]\n",
	       go->video_dev->name, video_device_node_name(go->video_dev));

	return 0;
}

void go7007_v4l2_remove(struct go7007 *go)
{
	mutex_lock(&go->hw_lock);
	if (go->streaming) {
		go->streaming = 0;
		go7007_stream_stop(go);
		if (go->streaming_file) {
			struct go7007_file *gofh = 
				go->streaming_file->private_data;
			videobuf_stop(&gofh->vb_vidq);
		}
	}
	mutex_unlock(&go->hw_lock);
	if (go->video_dev)
		video_unregister_device(go->video_dev);
	v4l2_device_unregister(&go->v4l2_dev);
}
